#include <stdio.h>
#include <boost/container/list.hpp>
#include <boost/container/map.hpp>
#include <boost/foreach.hpp>

//
template<unsigned B, unsigned N>
class SampleAlloc2 {
public:
	SampleAlloc2() {
		for (unsigned i = 0; i < N-1; ++i)
			buf_[i].link = &buf_[i+1];
		buf_[N-1].link = NULL;
		root_ = &buf_[0];
	}
	void* allocate(unsigned n) {
		assert(n == 1 && root_);
		Link*	p = root_;
		root_ = p->link;
		return p;
	}
	void  deallocate(void* t, unsigned n) {
		if (t) {
			assert(n == 1);
			Link* 	p = reinterpret_cast<Link*>(t);
			p->link	  = root_;
			root_	  = p;
		}
	}
private:
	union Link {
		Link*	link;
		char	b[B];
	};
private:
	Link*		root_;
	Link		buf_[N];
};

enum { ELM_SIZE = 32 };	// Reĩm[h܂񂾗vf̃oCg.
enum { MAX_SIZE = 16 };
template<typename T>
class SampleAllocator2 : public SampleAlloc2<ELM_SIZE,MAX_SIZE> {
	typedef SampleAlloc2<ELM_SIZE,MAX_SIZE> base_type;
public:
	typedef T value_type;
	SampleAllocator2() {BOOST_STATIC_ASSERT(sizeof(T) <= ELM_SIZE);}

	T* allocate(unsigned n) { return reinterpret_cast<T*>(base_type::allocate(n)); }
	void  deallocate(T* t, unsigned n) { base_type::deallocate(t, n); }
	bool operator == (SampleAllocator2 const & ) const { return false; }
	bool operator != (SampleAllocator2 const & ) const { return true; }
};

static void sampleList()
{
	typedef SampleAllocator2<int>	Allocator;
	typedef boost::container::list<int, Allocator >	List;
	List	li1;
	for (int i = 0; i < 4; ++i)
		li1.insert(li1.begin(), i);
	BOOST_FOREACH(int n, li1)
		printf("%d\n", n);
	List	li2 = li1;
	assert(std::equal(li1.cbegin(), li1.cend(), li2.cbegin()));
	List	li3;
	li3 = li2;
	assert(li3 == li2);
}

struct cstr_less {
	bool operator()(const char* l, const char* r) const { return strcmp(l,r) < 0; }
};

static void sampleMap()
{
	typedef std::pair<char const* const,int>	smpPair_t;
	typedef SampleAllocator2<smpPair_t>	Allocator;
	typedef boost::container::map<const char*, int, cstr_less, Allocator >	Map;
	Map		map1;
	map1["xyz"] = 1, map1["abc"] = 2, map1["lmn"] = 3;
	BOOST_FOREACH(smpPair_t p , map1)
		printf("%s %d\n", p.first, p.second);
	Map		map2 = map1;
	assert(std::equal(map1.cbegin(), map1.cend(), map2.cbegin()));
	Map		map3;
	map3	= map2;
	assert(map3 == map2);
}

void sample2()
{
	sampleList();
	sampleMap();
}
